#!/usr/bin/env python3
# coding: utf-8
from __future__ import print_function, unicode_literals

import time
import threading
from ripcap import Ripcap, Tcpdump, sys

"""
srvprof.py: guessing at bottlenecks through tcp profiling

monitors tcp sessions, displays time-to-first-byte for rx/tx and
  indicates which side was blocking (server/us or the client/remote)

screenshot: https://ocv.me/dev/srvprof.png
     video: https://ocv.me/dev/srvprof.webm

THE COLORS WHAT DO THEY MEAN
 * red: socket hasn't been accepted by server yet (eg. fail2ban)
 * yellow: active connection
 * cyan: either server or client sent FIN (shutdown)
 * gray: terminated connection

summary of the columns:
 * remote-ip/port: the non-server side of the connection
 * bytes.Rx/Tx: total bytes received by server / sent to client
 * #pkt.Rx/Tx: number of packets received / sent
 * T.conn: time until the server accepted the TCP handshake
 * 1st.Rx/Tx: time until the first byte was received/sent
     (starts counting when the TCP handshake is complelted)
 * idle-Rx/Tx: time spent processing data / preparing a reply
     on client-side (Rx) or server-side (Tx)
 * T.life: total life of the connection until it was closed

dependencies:
 * https://ocv.me/dev/ripcap.py
 * tcpdump
 * python2/3 or pypy2/3
"""
__version__ = (0, 7)
__date__ = "2020-10-08"
__author__ = "ed <irc.rizon.net>"
__url__ = "https://ocv.me/dev/?srvprof.py"
__credits__ = ["stackoverflow.com"]
__license__ = "MIT"
__copyright__ = 2020


PY2 = False
try:
    from queue import Queue
except:
    PY2 = True
    from Queue import Queue


def termsize():
    import os

    env = os.environ

    def ioctl_GWINSZ(fd):
        try:
            import fcntl, termios, struct, os

            cr = struct.unpack("hh", fcntl.ioctl(fd, termios.TIOCGWINSZ, "1234"))
        except:
            return
        return cr

    cr = ioctl_GWINSZ(0) or ioctl_GWINSZ(1) or ioctl_GWINSZ(2)
    if not cr:
        try:
            fd = os.open(os.ctermid(), os.O_RDONLY)
            cr = ioctl_GWINSZ(fd)
            os.close(fd)
        except:
            pass
    if not cr:
        try:
            cr = (env["LINES"], env["COLUMNS"])
        except:
            cr = (25, 80)
    return int(cr[1]), int(cr[0])


def humanize(v):
    if v < 100 * 1000 * 1000:
        return v

    suffix = ""
    for unit in ["k", "m"]:
        if abs(v) >= 1000:
            v /= 1000.0
            suffix = unit
    width = 0
    if v < 1000:
        width = 3
    if v < 100:
        width = 4
    if v < 10:
        width = 5
    return "{v:.{w}f}{s}".format(v=v, w=width, s=suffix)


class Conn(object):
    def __init__(self, t0, src_ip, src_port):
        # time the first SYN arrived at us
        self.t0 = t0

        # dst is implicit (us), only track src (remote)
        self.ip = src_ip
        self.port = src_port

        # time we (dst) sent our SYN-ACK accepting the socket
        self.estab = None

        # time src (remote) or dst (us) sent FIN
        self.fin_src = None
        self.fin_dst = None

        # time first data-packet was rx from remote / tx to remote
        self.first_rx = None
        self.first_tx = None

        # time of most recent packet
        self.last_rx = None
        self.last_tx = None

        # time $dir wasted before it sent the next packet:
        # -- rx then tx counts as tx idle (we were blocking)
        # -- tx then rx counts as rx idle (remote was blocking)
        # -- rx then another rx counts as rx idle (remote was blocking)
        self.idle_rx = 0
        self.idle_tx = 0

        # num packets sent/received
        self.num_rx = 0
        self.num_tx = 0

        # total bytes sent/received
        self.byte_rx = 0
        self.byte_tx = 0


def painter(conns, mtx):
    hist = []
    prev = {}
    height_ctr = 4321
    banner = (
        "\033[0m"
        + r"--remote-ip--    port | bytes.Rx bytes.Tx |  #pkt.Rx  #pkt.Tx |   T.conn   1st.Rx   1st.Tx |  idle-Rx  idle-Tx   T.life"
    )
    while True:
        height_ctr += 1
        if height_ctr > 10:
            height_ctr = 0
            height = termsize()[1]

        curr = {}
        pends = []
        actives = []
        fins = []
        now = time.time()
        with mtx:
            rm = []
            for k, c in conns.items():
                estab = c.estab or now
                if c.fin_src and c.fin_dst:
                    ended = max(c.fin_src, c.fin_dst)
                    rm.append(k)
                else:
                    ended = now

                ln = "{:15} {:5} | {:8} {:8} | {:8} {:8} | {:8.3f} {:8.3f} {:8.3f} | {:8.3f} {:8.3f} {:8.3f}".format(
                    c.ip,
                    c.port,
                    # |
                    humanize(c.byte_rx),
                    humanize(c.byte_tx),
                    # |
                    humanize(c.num_rx),
                    humanize(c.num_tx),
                    # |
                    estab - c.t0,
                    (c.first_rx or estab - 1) - estab,
                    (c.first_tx or estab - 1) - estab,
                    # |
                    c.idle_rx,
                    c.idle_tx,
                    ended - c.t0,
                )
                curr[k] = ln
                if not c.estab:
                    pends.append("\033[01;37;41m" + ln)
                elif not c.fin_src and not c.fin_dst:
                    actives.append("\033[01;33;40m" + ln)
                else:
                    fins.append("\033[0;36m" + ln)
            
            for k in rm:
                del conns[k]

        # take all conns that were removed just now
        # and persist them in hist[]
        for k, v in prev.items():
            if k not in curr:
                hist.append(v)

        prev = curr
        if len(hist) > 1024 + 128:
            hist = hist[-1024:]

        # draw it
        header = "\033[H\033[0m{}\n".format(banner)
        lines = hist + fins + actives + pends + [banner + " \033[J"]
        scrn = header + "\033[K\n".join(lines[-(height - 1) :])
        print(scrn, end="")
        time.sleep(0.03)


def eater(que):
    mtx = threading.Lock()
    conns = {}

    t = threading.Thread(
        target=painter,
        args=(
            conns,
            mtx,
        ),
    )
    t.daemon = True
    t.start()

    while True:
        ts, fr = que.get()
        proto = fr.header2["type"]
        if proto != 6:
            # not tcp
            continue

        flags = fr.header2["flags"]
        is_syn = flags & 0x02 != 0
        is_ack = flags & 0x10 != 0
        is_fin = flags & 0x01 != 0
        is_rst = flags & 0x04 != 0

        k1 = "{}-{}".format(fr.src_ip, fr.src_port)
        k2 = "{}-{}".format(fr.dst_ip, fr.dst_port)

        with mtx:
            if is_rst and False:
                # thats a full-duplex termination yeah? TODO chk
                try:
                    conns[k1].fin_src = ts
                    conns[k1].fin_dst = ts
                except:
                    # dont think we can end up here but it's 2:36am
                    conns[k2].fin_src = ts
                    conns[k2].fin_dst = ts

                continue

            if is_fin:
                both = False
                if k1 in conns:
                    k = k1
                    conns[k].fin_src = ts
                    if conns[k].fin_dst:
                        both = True
                elif k2 in conns:
                    k = k2
                    conns[k].fin_dst = ts
                    if conns[k].fin_src:
                        both = True

                #if both:
                #    del conns[k]

                continue

            if is_syn:
                if not is_ack:
                    # new connection from remote
                    conns[k1] = Conn(ts, fr.src_ip, fr.src_port)
                else:
                    # syn-ack (us accepting the connection),
                    # set established timestamp
                    # TODO verify syn/ack seq
                    c = conns[k2]
                    c.estab = ts

                continue

            if not fr.data:
                continue

            if k1 in conns:
                k = k1
                tx = False
                c = conns[k]
                c.num_rx += 1
                c.byte_rx += len(fr.data)
                if c.first_rx is None:
                    c.first_rx = ts
            elif k2 in conns:
                k = k2
                tx = True
                c = conns[k]
                c.num_tx += 1
                c.byte_tx += len(fr.data)
                if c.first_tx is None:
                    c.first_tx = ts
            else:
                # we didn't catch the beginning of this
                continue

            last_act = max(c.last_rx or c.estab, c.last_tx or c.estab)
            if tx:
                c.idle_tx += ts - last_act
                c.last_tx = ts
            else:
                c.idle_rx += ts - last_act
                c.last_rx = ts


def main():
    try:
        iface, cap_filter = sys.argv[1:]
    except:
        print()
        print('need arg 1: interface to monitor')
        print('need arg 2: tcpdump filter string')
        print()
        print('example:')
        print('  srvprof.py eth0 "port 443"')
        print()
        sys.exit(1)

    print("\033[H\033[32m\033[Jhello world\033[0m")

    que = Queue(1024)
    t = threading.Thread(target=eater, args=(que,))
    t.daemon = True
    t.start()

    p = Ripcap(Tcpdump(iface, cap_filter))
    for x in p.parse_pcap():
        que.put(x)


if __name__ == "__main__":
    main()


# socat tcp-listen:4321,fork,reuseaddr - | pv -i 0.1 >/dev/null
#
# while true; do pv /dev/zero -Ss $((1024*128)) -L 100K -i 0.1 | socat - tcp-connect:127.0.0.1:4321; sleep 1; done
# while true; do pv /dev/zero -Ss $((1024*128)) -L 50K -i 0.1 | socat - tcp-connect:127.0.0.1:4321; sleep 0.2; done
# while true; do (sleep 0.1; pv /dev/zero -Ss $((1024*128)) -L 50K -i 0.1) | socat - tcp-connect:127.0.0.1:4321; sleep 0.2; done
#
# clear; printf '\n  this window does nothing\n   just makes the window above smaller\n    so it hits the end faster\n\n  most of the windows on the right send zeros over tcp\n   the bottom right window is a tcp server that discards all of it\n\n  the window above this one uses wireshark to monitor tcp connections\n   and tracks how quickly they reached first-byte rx/tx\n    and also which side of the connection was pissing away time\n\n'; read

